/* agent.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "agent.h"
#include "error.h"
#include "stype.h"
#include "atype.h"
#include "multilist.h"

// ------------------------------------------------------------------
// ȥޥ͡

AgentManager *agent_manager_open(int size)
{
  AgentManager *am;
  int i, *p;

  am = (AgentManager *)malloc(sizeof(AgentManager));
  if( am == NULL )
    error(MALLOC_FAILED,"agent_manager_open 1");
  am->list = (Agent *)malloc(sizeof(Agent) * size);
  if( am->list == NULL )
    error(MALLOC_FAILED,"agent_manager_open 2");
  if( (am->table = (int *)malloc(sizeof(int) * size)) == NULL )
    error(MALLOC_FAILED,"agent_manager_open 3");
  am->max = size;
  am->pivot = 0;
  p = am->table;
  for( i = 0; i < size; p += 1, i += 1 ){
    *p = i;
  }
  
  am->actlist = numstack_open(size);
  am->slplist = numstack_open(size);
  am->tmplist = numstack_open(size);
  am->angellist = numstack_open(size);

  agent_manager_clear(am);

  return am;
}



void agent_manager_clear(AgentManager *am)
{
  am->last_name = 0;
  am->pivot = 0;

  numstack_clear(am->actlist);
  numstack_clear(am->slplist);
  numstack_clear(am->tmplist);
  numstack_clear(am->angellist);
}



void agent_manager_close(AgentManager *am)
{
  free(am->list);
  free(am->table);
  numstack_close(am->actlist);
  numstack_close(am->slplist);
  numstack_close(am->tmplist);
  numstack_close(am->angellist);
  free(am);
}


void agent_manager_dump(AgentManager *am, FILE *fp)
{
  int i, j, used, flag;

  fprintf(fp, "  AgentManager (max = %d)\n", am->max);
  fprintf(fp, "  table: ");

  used = -1;
  for( i = 0; i < am->pivot; i += 1 ){
    fprintf(fp, "a%d ", am->table[i]);
    if( am->table[i] > used ){
      used = am->table[i];
    }
  }
  fprintf(fp, "\n  agent: ");
  for( i = 0; i <= used; i += 1 ){
    flag = 0;
    for( j = 0; j < am->pivot; j += 1 )
      if( am->table[j] == i )
	flag = 1;
    if( flag )
      fprintf(fp, "a%d ", i);
    else
      fprintf(fp, "   ");
  }

  fprintf(fp, "\n  name:  ");
  for( i = 0; i <= used; i += 1 ){
    flag = 0;
    for( j = 0; j < am->pivot; j += 1 )
      if( am->table[j] == i )
	flag = 1;
    if( flag )
      fprintf(fp, "n%d ", am->list[i].name);
    else
      fprintf(fp, "   ");
  }
  fprintf(fp,"\n");
}


void agent_manager_test(AgentManager *am)
{
  int *chk,i;

  chk = (int *)malloc(sizeof(int)*am->max);
  for( i = 0; i < am->max; i += 1 ){
    chk[i] = 0;
  }

  for( i = 0; i < am->max; i += 1 ){
    chk[am->table[i]] = 1;
  }

  for( i = 0; i < am->max; i += 1 ){
    if( chk[i] == 0 )
      error(UNKNOWN,"agent_manager_test");
  }
}



// ==================================================================
// ȸ̤


// ------------------------------------------------------------------
// ˥

void agent_print_detail(AgentManager *am, int agt_idx, FILE *fp)
{
  Agent *agt;
  int i;

  agt = &am->list[agt_idx];
  fprintf(fp,"-----   agt%-3d ----------------------------------------\n",agt->name);
  fprintf(fp," = agt%d,  = %s, ⡼ = %s, HP = %d\n",
	  agt->parent,
	  (agt->age == CHILD) ? "CHILD" : "ADULT",
	  (agt->mode == FULLMODE) ? "FULLMODE" : (agt->mode == HALFMODE) ? "HALFMODE" : (agt->mode == SLEEPMODE) ? "SLEEPMODE" : (agt->mode == HELLMODE) ? "HELLMODE" : "HEAVENMODE",
	  agt->hp);
  fprintf(fp,"%s = %d, DEPTH = %d AWAKE_TIME = %d\n",
	  (agt->loc_type == LOC_NODE) ? "node" : "arc",
	  agt->loc_idx, agt->depth, agt->awake_time);
  fprintf(fp,"Refered for SYN: mll%2d     Pop & Refered for SYN: mll%2d\n", agt->rfrd_syn, agt->prfrd_syn);
  fprintf(fp,"Refered for SEM: mll%2d     Pop & Refered for SEM: mll%2d\n", agt->rfrd_sem, agt->prfrd_sem);
  fprintf(fp,"mystack: ");
  for( i = 0 ; i < agt->mysp; i += 1 ){
    fprintf(fp,"[%d]", agt->mystack[i]);
  }
  fprintf(fp,"\n");
  fprintf(fp,"-------------------------------------------------------\n");
}


// ------------------------------------------------------------------ 
// ȯȾ
// ʤΥ꡼ϡѤΤǡȤϥǥ
// ϤȤ

// ȯ
int agent_new(AgentManager *am)
{
  int ans, mypivot;
  Agent *agt;

  // ----- ȴ -----
  mypivot = am->pivot++;
  if( mypivot >= am->max )
    error(OVER_FLOWED,"Too many agents.");
  ans = am->table[mypivot];
  agt = &am->list[ans];
  agt->mypivot = mypivot;

  // ----- Ⱦ -----
  agt->name = am->last_name++;
  agt->parent = ans;
  agt->age = ADULT;
  agt->mode = FULLMODE;
  agt->hp = DEFAULT_HP;
  agt->loc_type = LOC_NODE;
  agt->loc_idx = 0;
  agt->depth = 1;
  agt->rfrd_syn = 0;
  agt->rfrd_sem = 0;
  agt->prfrd_syn = 0;
  agt->prfrd_sem = 0;
  agt->mysp = 0;
  memset(agt->cnt_memory,0,sizeof(int)*AGENT_CNT_MEMORY_SIZE);
  memset(agt->loc_memory,0,sizeof(int)*AGENT_LOC_MEMORY_SIZE);
  agt->awake_time = 0;

  return ans;
}



// Ҷ򻺤
int agent_breed(AgentManager *am, int parent_idx)
{
  int child_idx, mypivot;
  Agent *child, *parent;

  // ----- ȴ -----
  mypivot = am->pivot++;
  if( mypivot >= am->max )
    error(OVER_FLOWED,"Too many agents.");
  child_idx = am->table[mypivot];
  child = &am->list[child_idx];
  child->mypivot = mypivot;

  // ----- Ⱦ -----
  parent = &am->list[parent_idx];

  child->name = am->last_name++;
  child->parent = parent_idx;
  child->age = CHILD;
  child->mode = parent->mode;
  child->hp = DEFAULT_HP;
  child->loc_type = parent->loc_type;
  child->loc_idx = parent->loc_idx;
  child->depth = parent->depth;
  child->rfrd_syn = parent->rfrd_syn;
  child->rfrd_sem = parent->rfrd_sem;
  child->prfrd_syn = parent->prfrd_syn;
  child->prfrd_sem = parent->prfrd_sem;

  // child->cnt_memory  // ĹȤˡ
  // child->loc_memory  // ʬΥ
  // child->mystack     // Ĥ褦ˤʤ롣
  // child->mysp

  child->awake_time = parent->awake_time;

  return child_idx;
}


// ͤˤʤ, ͤä顢Τޤޡ
void agent_growup(AgentManager *am, int agt_idx)
{
  Agent *agt, *parent;
  int i, psp;

  agt = &am->list[agt_idx];
  if( agt->age == ADULT )
    return;

  parent = &am->list[agt->parent];
  agt->age = ADULT;
  psp = parent->mysp;
  for( i = 0; i < psp; i += 1 )
    agt->mystack[i] = parent->mystack[i];
  agt->mysp = psp;
  memcpy(agt->cnt_memory,parent->cnt_memory,sizeof(int)*AGENT_CNT_MEMORY_SIZE);
  memcpy(agt->loc_memory,parent->loc_memory,sizeof(int)*AGENT_LOC_MEMORY_SIZE);

  // ɽ̲ᵭϿϸ
  // ̲ᵭϿϸ
}


// 
void agent_kill(AgentManager *am, int agt_idx)
{
  Agent *agt1, *agt2;
  int p;

  am->pivot -= 1;
  p = am->pivot;
  if( p < 0 )
    error(UNDER_FLOWED,"agent_kill");
  agt1 = &am->list[agt_idx];
  if( agt1->mypivot != p ){
    agt2 = &am->list[am->table[p]];
    agt2->mypivot = agt1->mypivot;
    am->table[agt1->mypivot] = am->table[p];
    am->table[p] = agt_idx;
  }
}


// ------------------------------------------------------------------
// ư
// 饤ŸʤΤǡإåե뻲

// ------------------------------------------------------------------
// 

// -----  -----

void agent_push(Agent *agt, int num)
{
  if( agt->mysp >= AGENT_MYSTACK_SIZE )
    error(OVER_FLOWED,"agent_push");
  agt->mystack[agt->mysp++] = num;
}


int agent_pop(Agent *agt)
{
  if( agt->mysp <= 0 )
    error(UNDER_FLOWED,"agent_pop");
  agt->mysp -= 1;
  return agt->mystack[agt->mysp];
}


// -----  -----
void agent_increase_counter(Agent *agt, int cntnum)
{
  if( cntnum > AGENT_CNT_MEMORY_SIZE )
    error(OVER_FLOWED,"agent_increase_counter");
  agt->cnt_memory[cntnum] += 1;
}


int agent_compare_counter(Agent *agt, int cntnum, int cndval, CmpType ct)
{
  if( cntnum > AGENT_CNT_MEMORY_SIZE )
    error(OVER_FLOWED,"agent_compare_counter");
  
  switch( ct ){
  case EQ:
    return( agt->cnt_memory[cntnum] == cndval );
  case NE:
    return( agt->cnt_memory[cntnum] != cndval );
  }
  error(UNKNOWN,"agent_compare_counter");
  return 0;
}

// ------ ȹ̡δϢ -----
void agent_store_local_result(AgentManager *am, Agent *agt, int rslt )
{
  // ϲ⤷ʤ
}


void agent_store_global_result(AgentManager *am, Agent *agt, int cnd_idx, int rslt )
{
  // ϲ⤷ʤ
}
