/*
 * eucmap.c
 * ʤ߷
 * 2005.7/12 by tokuhisa
 * first version 2003, April. by tishida
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "error.h"
#include "eucmap.h"

/*
 * EUCɤ򰵽̤
 *    IN:   unsigned char b1  ʸ1Х  a0-ff (a1-f4)
 *          unsigned char b2  ʸ2Х  a0-ff (a1-fe)
 *    OUT:  ʸ (0x0000  0x23FF)
 */

#define eucmap_encode(_x,_y) (unsigned int)(((_x) - 0xa0) * 96 + ((_y) - 0xa0))

/*
 * ʸɥӥåȥޥåΰ֤ȥӥåȥޥåפ
 *    IN:    code  ʸ
 *    OUT:   󥤥ǥ
 *           1Υӥåȥޥå
 */

#define eucmap_index(code) ((unsigned int)(code >> 5))
#define eucmap_partial_map(code) ((unsigned int)(1 << (code & 0x1f)))

/*
 * ӥåȥޥåפ
 *    IN:    ʸ
 *           ӥåȥޥåפν񤭹ΰ
 *    OUT:   ӥåȥޥå
 */

void eucmap_create_bitmap(unsigned char *str, int len, EucAllMap bitmap)
{
  unsigned char *p1, *p2;
  int i;
  unsigned short code;
  unsigned int map;
  int indx;

  p1 = str;
  p2 = str+1;
  for( i = 0; i < len; i += 2 ){
    code = eucmap_encode( *p1, *p2 );
    indx = eucmap_index( code );
    map = eucmap_partial_map( code );
    bitmap[indx] |= map;
    p1 += 2;
    p2 += 2;
  }
}


/*
 * Ψӥåȥޥåפ
 *    IN:    ӥåȥޥå
 *           񤭹ΰ
 *    OUT:   񤭹Υǥ
 *           񤭹߸Ŀ
 */

void eucmap_set_bitmap(EucAllMap bitmap, EucPrtMapBank *eb, int *idx, int *num)
{
  int i, cnt;

  *idx = eb->free;
  cnt = 0;
  for( i = 0 ; i < EUCMAP_WIDTH; i += 1 ){
    if( bitmap[i] != 0 ){
      if( eb->free == eb->max )
	error(OVER_FLOWED, "eucmap_set_bitmap");
      eb->pmap[eb->free].index = i;
      eb->pmap[eb->free++].pmap = bitmap[i];
      cnt += 1;
    }
  }
  *num = cnt;
}



/* ɽ */

void eucmap_dump_bitmap(EucAllMap bitmap, FILE *fp)
{
  int i;

  for( i = 0; i < EUCMAP_WIDTH; i += 1 ){
    if( bitmap[i] != 0 )
      printf("epm---:[%3d,0x%08x]\n",i,bitmap[i]);
  }
}



/*
 * ӥåȥޥåפΨӥåȥޥåפޤळȤǧ
 *    IN:    ӥåȥޥå
 *           Ψӥåȥޥåΰ
 *           ǥƬ
 *           Ŀ
 *    OUT:   0.. ޤޤʤ1..ޤ
 */

int eucmap_check_bitmap_including(EucAllMap bitmap, EucPrtMapBank *eb, int indx, int num)
{
  EucPrtMap *e;

  e = &eb->pmap[indx];
  for(; num > 0; e += 1, num -= 1)
    if( ((~bitmap[e->index]) & e->pmap) != 0 )
      return 0;
  return 1;
}


void eucmap_bank_clear(EucPrtMapBank *epb)
{
  epb->free = 0;
}


EucPrtMapBank *eucmap_bank_open(int size)
{
  EucPrtMapBank *epb;

  if( (epb = (EucPrtMapBank *)malloc(sizeof(EucPrtMapBank))) == NULL )
    error(MALLOC_FAILED,"eucmap_create_bank 1");
  if( (epb->pmap = (EucPrtMap *)malloc(sizeof(EucPrtMap) * size)) == NULL )
    error(MALLOC_FAILED,"eucmap_create_bank 2");
  epb->max = size;
  eucmap_bank_clear(epb);
  return epb;
}


void eucmap_bank_close(EucPrtMapBank *epb)
{
  free(epb->pmap);
  free(epb);
}


void eucmap_bank_dump(EucPrtMapBank *epb, FILE *fp)
{
  int i;

  fprintf(fp,"EucPrtMapBank  (%d/%d)\n",epb->free,epb->max);
  for( i = 0; i < epb->free; i += 1 )
    fprintf(fp,"epm%3d:[%3d,0x%08x]\n", i, epb->pmap[i].index, epb->pmap[i].pmap);
}


void eucmap_bank_save(EucPrtMapBank *epb, FILE *fp)
{
  fprintf(fp, "%d %d\n", epb->max, epb->free);
  fwrite(epb->pmap, sizeof(EucPrtMap), epb->free, fp);
}


EucPrtMapBank *eucmap_bank_load(FILE *fp)
{
  EucPrtMapBank *epb;

  if( (epb = (EucPrtMapBank *)malloc(sizeof(EucPrtMapBank))) == NULL )
    error(MALLOC_FAILED,"eucmap_bank_load 1");
  fscanf(fp,"%d %d\n", &epb->max, &epb->free);
  if( (epb->pmap = (EucPrtMap *)malloc(sizeof(EucPrtMap)*epb->max)) == NULL )
    error(MALLOC_FAILED,"eucmap_bank_load 2");
  fread(epb->pmap, sizeof(EucPrtMap), epb->free, fp);

  return epb;
}

void eucmap_bank_strip(EucPrtMapBank *epb)
{
  epb->max = epb->free + 1;
}


void eucmap_decode(int indx, unsigned int pmap)
{
  unsigned int base, code;
  unsigned char b1, b2;
  int cnt;

  base = indx << 5;
  for( cnt = 0; cnt < 32; cnt += 1 ){
    if( (pmap & (1 << cnt)) != 0 ){
      code = base | cnt;
      b1 = (code / 96) + 0xa0;
      b2 = (code % 96) + 0xa0;
      printf("(%c%c)\n",b1,b2);
    }
  }
}
