/*
 * inmorph.c ver.2
 * 2005.7/1 by tokuhisa
 * Ṳ̆򵭲뤿ι¤ѹ
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "error.h"
#include "scanner.h"
#include "inmorph.h"
#include "eucmap.h"

// ==================================================================

void inmorph_clear(InMorph *im)
{
  im->block[0] = '\0';

  im->mls_free = 0;
  im->mps_free = 0;
  im->mas_free = 0;
  im->mll_free = 0;

  multi2list_clear(im->mpl);
  multilist_clear(im->mal);
}


InMorph *inmorph_open(int s1, int s2, int s3, int s4, int s5, int s6, int s7)
{
  InMorph *im;

  if( (im = (InMorph *)malloc(sizeof(InMorph))) == NULL )
    error(MALLOC_FAILED,"inmorph_open 1");
  if( (im->block = (char *)malloc(sizeof(char) * s1)) == NULL )
    error(MALLOC_FAILED,"inmorph_open 2");
  im->block_max = s1;
  if( (im->mls = (MorphLineStr *)malloc(sizeof(MorphLineStr) * s2)) == NULL )
    error(MALLOC_FAILED,"inmorph_open 3");
  im->mls_max = s2;
  if( (im->mps = (MorphPosStr *)malloc(sizeof(MorphPosStr) * s3)) == NULL )
    error(MALLOC_FAILED,"inmorph_open 4");
  im->mps_max = s3;
  if( (im->mas = (MorphAttrStr *)malloc(sizeof(MorphAttrStr) * s4)) == NULL )
    error(MALLOC_FAILED,"inmorph_open 5");
  im->mas_max = s4;

  if( (im->mll = (MorphLineLink *)malloc(sizeof(MorphLineLink) * s5)) == NULL )
    error(MALLOC_FAILED,"inmorph_open 6");
  im->mll_max = s5;

  im->mpl = multi2list_open(s6);
  im->mal = multilist_open(s7);
  inmorph_clear(im);

  return im;
}



void inmorph_close(InMorph *im)
{
  free(im->block);
  free(im->mls);
  free(im->mps);
  free(im->mas);
  free(im->mll);
  multi2list_close(im->mpl);
  multilist_close(im->mal);
  free(im);
}


void putsn(FILE *fp, char *str, int len)
{
  for( ; len > 0; len -= 1 )
    fputc(*str++, fp);
}


void inmorph_dump(InMorph *im, FILE *fp)
{
  int i, cnt;
  int lst1, lst2;
  ML2Node *n;

  fprintf(fp, "InMorph (block = %d, mls = %d/%d, mps = %d/%d, mas = %d/%d, mll = %d/%d)\n",
	  im->block_max, im->mls_free, im->mls_max,
	  im->mps_free, im->mps_max, im->mas_free, im->mas_max,
	  im->mll_free, im->mll_max);

  fprintf(fp, "%s\nMorpLineStr\n", im->block);
  for( i = 0; i < im->mls_free; i += 1 ){
    fprintf(fp, "  ls%-2d:[%c,", i, im->mls[i].border);
    putsn(fp, im->mls[i].raw, im->mls[i].raw_len);
    fprintf(fp, "]\n");
  }

  fprintf(fp,"MorphPosStr\n");
  for( i = 0; i < im->mps_free; i += 1 ){
    fprintf(fp, "  ps%-2d:[",i);
    putsn(fp,im->mps[i].pcode,4); fputc(',',fp);
    putsn(fp,im->mps[i].std,im->mps[i].std_len); fputc(',',fp);
    putsn(fp,im->mps[i].chs,im->mps[i].chs_len); fputs("]\n",fp);
  }

  fprintf(fp,"MorphAttrStr\n");
  cnt = 0;
  for( i = 0; i < im->mas_free; i += 1 ){
    fprintf(fp, "  as%-2d:[",i);
    putsn(fp,im->mas[i].cat,im->mas[i].cat_len); putc(':',fp);
    putsn(fp,im->mas[i].val,im->mas[i].val_len); fputs("]",fp);
    cnt += 1;
    if( cnt % 4 == 0 )
      fprintf(fp,"\n");
  }

  fprintf(fp,"\nMorphLineLink\n");
  for( i = 0; i < im->mll_free; i += 1 ){
    fprintf(fp, "  ll%-2d:[(%s),ls%d,",i,
	    im->mll[i].cnt == CONT ? "" : im->mll[i].cnt == NIL ? "nl" : "",
	    im->mll[i].mls_idx);
    if( im->mll[i].mpl_head == 0 )
      fprintf(fp,"----]\n");
    else 
      fprintf(fp,"pl%d]\n", im->mll[i].mpl_head);
  }

  fprintf(fp,"MorphPosLink\n");
  for( i = 0; i < im->mll_free; i += 1 ){
    lst1 = im->mll[i].mpl_head;
    if( lst1 != 0 ){
      fprintf(fp, "  ll%-2d -> pl%-2d:[",i,lst1);
      while( !multi2list_empty(im->mpl,lst1) ){
	n = multi2list_car(im->mpl,lst1);
	fprintf(fp," (ps%d,al%d)", n->val1, n->val2);
	lst1 = multi2list_cdr(im->mpl,lst1);
      }
      fprintf(fp," ]\n");
    }
  }

  fprintf(fp,"MorphAttrLink\n");
  for( i = 0; i < im->mll_free; i += 1 ){
    lst1 = im->mll[i].mpl_head;
    if( lst1 != 0 ){
      while( !multi2list_empty(im->mpl,lst1) ){
	n = multi2list_car(im->mpl,lst1);
	lst2 = n->val2;
	if( lst2 != 0 ){
	  fprintf(fp,"  ll%-2d -> pl%-2d -> al%-2d:[",i,lst1,lst2);
	  while( !multilist_empty(im->mal,lst2) ){
	    fprintf(fp," (as%d)", multilist_car(im->mal,lst2));
	    lst2 = multilist_cdr(im->mal,lst2);
	  }
	  fprintf(fp," ]\n");
	}
	lst1 = multi2list_cdr(im->mpl,lst1);
      }
    }
  }
}


int parse_MORPH(InMorph *, int);

void inmorph_load(InMorph *im, FILE *fp)
{
  char line[BUFSIZ];
  int len;

  while( fgets(line,BUFSIZ,fp) != NULL )
    strcat(im->block, line);

  len = parse_MORPH(im, 0);
  if( (len == 0) || (im->block[len] != '\0') )
    error(SYNTAX_ERROR,"morph file");
}


void inmorph_set_eucbitmap(InMorph *im, EucAllMap bitmap)
{
  int i, lines;

  eucmap_bitmap_clear(bitmap);
  lines = im->mls_free;
  for( i = 0; i < lines; i += 1 ){
    eucmap_create_bitmap( im->mls[i].raw, im->mls[i].raw_len, bitmap );
  }
  lines = im->mps_free;
  for( i = 0; i < lines; i += 1 ){
    eucmap_create_bitmap( im->mps[i].std, im->mps[i].std_len, bitmap );
    eucmap_create_bitmap( im->mps[i].chs, im->mps[i].chs_len, bitmap );
  }
}




int inmorph_count_mll(InMorph *im, int ref)
{
  int ans;

  ans = 1;
  while( im->mll[ref++].cnt == CONT )
    ans += 1;
  return ans;
}




void inmorph_print_easily(InMorph *im, int ref, FILE *fp)
{
  int i, j;
  int lines;

  lines = im->mls_free;
  fprintf(fp, "ʸ");
  for( i = 0; i < lines; i += 1 ){
    for( j = 0; j < im->mls[i].raw_len; j += 1 )
      fprintf(fp, "%c", im->mls[i].raw[j]);
    fprintf(fp, " ");
  }

  fprintf(fp, "\n        ");
  for( i = 0; i < ref; i += 1 ){
    for( j = 1; j < im->mls[i].raw_len; j += 1 )
      fprintf(fp, " ");
    fprintf(fp, "  ");
  }
  fprintf(fp, "\n");
}





// ==================================================================

int parse_SEMELEM(InMorph *im, int p, int mal_head)
{
  int sum, len;
  MorphAttrStr *mas;

  if( im->mas_free >= im->mas_max )
    error(OVER_FLOWED,"parse_SEMELEM");

  mas = &im->mas[im->mas_free];
  if( (sum = scan_LALPHA(im->block, p)) > 0 )
    if( (len = scan_LALPHA(im->block, p + sum)) > 0 ){
      sum += len;
      mas->cat = &im->block[p];
      mas->cat_len = sum;
      if( (len = scan_COLON(im->block, p + sum)) > 0 ){
	sum += len;
	if( (len = scan_NUMALPHAS(im->block, p + sum)) > 0 ){
	  mas->val = &im->block[p+sum];
	  mas->val_len = len;
	  multilist_enqueue(im->mal, mal_head, im->mas_free++);
	  return sum + len;
	}
      }
    }
  return 0;
}



int parse_SOPTION(InMorph *im, int p, int mal_head)
{
  int sum, len;
  
  if( (sum = scan_ML(im->block, p)) > 0 ){
    if( (len = parse_SEMELEM(im, p + sum, mal_head)) > 0 ) {
      sum += len;
      while( (len = scan_COMMA(im->block, p + sum)) > 0 ) {
	sum += len;
	if( (len = parse_SEMELEM(im, p + sum, mal_head)) > 0 ){
	  sum += len;
	} else
	  return 0;
      }
      if( (len = scan_MR(im->block, p + sum)) > 0 )
	return sum + len;
    }
  }
  return 0;
}



int parse_PCODE(InMorph *im, int p)
{
  int sum, len;

  if( (sum = scan_NUMALPHA(im->block, p)) > 0 )
    if( (len = scan_NUMALPHA(im->block, p + sum)) > 0 ){
      sum += len;
      if( (len = scan_NUMALPHA(im->block, p + sum)) > 0 ){
	sum += len;
	if( (len = scan_NUMALPHA(im->block, p + sum)) > 0 ){
	  return sum + len;
	}
      }
    }
  return 0;
}



int parse_POS(InMorph *im, int p, int mpl_head)
{
  int sum, len, mal_head;
  MorphPosStr *mps;

  if( im->mps_free >= im->mps_max )
    error(OVER_FLOWED,"parse_POS");
  mps = &im->mps[im->mps_free];
  mal_head = multilist_new_head(im->mal);

  if( (sum = scan_PL(im->block, p)) > 0 ){
    sum += scan_MARK(im->block, p + sum);	// ɤФ
    if( (len = parse_PCODE(im, p + sum)) > 0 ){
      mps->pcode = &im->block[p + sum];
      sum += len;
      if( (len = scan_PR(im->block, p + sum)) > 0 ){
	mps->std_len = 0;
	mps->chs_len = 0;
	multi2list_enqueue(im->mpl, mpl_head, im->mps_free++, 0);
	multilist_close_head(im->mal,mal_head);
	return sum + len;
      }
      if( (len = scan_COMMA(im->block, p + sum)) > 0 ){
	sum += len;
	if( (len = scan_COMMA(im->block, p + sum)) > 0 ){
	  sum += len;
	  if( (len = scan_EUCS(im->block, p + sum)) > 0 ){
	    mps->chs = &im->block[p + sum];
	    mps->chs_len = len;
	    sum += len;
	    if( (len = scan_PR(im->block, p + sum)) > 0 ){
	      multi2list_enqueue(im->mpl, mpl_head, im->mps_free++, 0);
	      multilist_close_head(im->mal,mal_head);
	      return sum + len;
	    } else if( (len = scan_COMMA(im->block, p + sum)) > 0 ){
	      sum += len;
	      if( (len = parse_SOPTION(im, p + sum, mal_head)) > 0 ){
		sum += len;
		if( (len = scan_PR(im->block, p + sum)) > 0 ){
		  multi2list_enqueue(im->mpl, mpl_head, im->mps_free++, mal_head);
		  return sum + len;
		}
	      }
	    }
	  } else if( (len = parse_SOPTION(im, p + sum, mal_head)) > 0 ){
	    sum += len;
	    if( (len = scan_PR(im->block, p + sum)) > 0 ){
	      multi2list_enqueue(im->mpl, mpl_head, im->mps_free++, mal_head);
	      return sum + len;
	    }
	  }
	} else if( (len = scan_COMMA(im->block, p + sum)) > 0 ){
	  sum += len;
	  if( (len = parse_SOPTION(im, p + sum, mal_head)) > 0 ){
	    sum += len;
	    if( (len = scan_PR(im->block, p + sum)) > 0 ){
	      multi2list_enqueue(im->mpl, mpl_head, im->mps_free++, mal_head);
	      return sum + len;
	    }
	  }
	}
	if( (len = scan_EUCS(im->block, p + sum)) > 0 ){
	  mps->std = &im->block[p + sum];
	  mps->std_len = len;
	  sum += len;
	  if( (len = scan_PR(im->block, p + sum)) > 0 ){
	    mps->chs_len = 0;
	    multi2list_enqueue(im->mpl, mpl_head, im->mps_free++, 0);
	    multilist_close_head(im->mal,mal_head);
	    return sum + len;
	  }
	  if( (len = scan_COMMA(im->block, p + sum)) > 0 ){
	    sum += len;
	    if( (len = scan_EUCS(im->block, p + sum)) > 0 ){
	      mps->chs = &im->block[p + sum];
	      mps->chs_len = len;
	      sum += len;
	      if( (len = scan_PR(im->block, p + sum)) > 0 ){
		multi2list_enqueue(im->mpl, mpl_head, im->mps_free++, 0);
		multilist_close_head(im->mal,mal_head);
		return sum + len;
	      } else if( (len = scan_COMMA(im->block, p + sum)) > 0 ){
		sum += len;
		if( (len = parse_SOPTION(im, p + sum, mal_head)) > 0 ){
		  sum += len;
		  if( (len = scan_PR(im->block, p + sum)) > 0 ){
		    multi2list_enqueue(im->mpl, mpl_head, im->mps_free++, mal_head);
		    return sum + len;
		  }
		}
	      }
	    } else if( (len = parse_SOPTION(im, p + sum, mal_head)) > 0 ) {
	      sum += len;
	      if( (len = scan_PR(im->block, p + sum)) > 0 ){
		mps->chs_len = 0;
		multi2list_enqueue(im->mpl, mpl_head, im->mps_free++, mal_head);
		return sum + len;
	      }
	    }
	  }
	} else if( (len = parse_SOPTION(im, p + sum, mal_head)) > 0 ){
	  sum += len;
	  if( (len = scan_PR(im->block, p + sum)) > 0 ){
	    mps->std_len = 0;
	    mps->chs_len = 0;
	    multi2list_enqueue(im->mpl, mpl_head, im->mps_free++, mal_head);
	    return sum + len;
	  }
	}
      }
    }
  }
  return 0;
}





int parse_LINENIL(InMorph *im, int p)
{
  int sum, len;
  MorphLineStr *mls;
  MorphLineLink *mll;

  if( (sum = scan_NUMS(im->block, p)) > 0 )
    if( (len = scan_DOT(im->block, p + sum)) > 0 ){
      sum += len;
      if( (len = scan_SPC(im->block, p + sum)) > 0 ){
	sum += len;
	if( (len = scan_BORDER(im->block, p + sum)) > 0 ){
	  if( im->mls_free >= im->mls_max )
	    error(OVER_FLOWED,"parse_LINENIL");
	  mls = &im->mls[im->mls_free];
	  mls->border = im->block[p + sum];
	  sum += len;
	  if( (len = scan_NIL(im->block, p + sum)) > 0 ){
	    sum += len;
	    if( (len = scan_RET(im->block, p + sum)) > 0 ){
	      mls->raw = NULL;
	      mls->raw_len = 0;
	      mll = &im->mll[im->mll_free];
	      mll->cnt = NIL;
	      mll->mls_idx = im->mls_free;
	      mll->mpl_head = 0;
	      im->mls_free += 1;
	      im->mll_free += 1;
	      return sum + len;
	    }
	  }
	}
      }
    }
  return 0;
}




int parse_LINE(InMorph *im, int p)
{
  int sum, len;
  MorphLineStr *mls;
  MorphLineLink *mll;

  if( (sum = scan_NUMS(im->block, p)) > 0 )
    if( (len = scan_DOT(im->block, p + sum)) > 0 ){
      sum += len;
      if( (len = scan_SPC(im->block, p + sum)) > 0 ){
	sum += len;
	if( (len = scan_BORDER(im->block, p + sum)) > 0 ){
	  if( im->mls_free >= im->mls_max )
	    error(OVER_FLOWED,"parse_LINE 1");
	  mls = &im->mls[im->mls_free];
	  mls->border = im->block[p + sum];
	  sum += len;
	  if( (len = scan_EUCS(im->block, p + sum)) > 0 ){
	    mls->raw = &im->block[p+sum];
	    mls->raw_len = len;
	    sum += len;
	    if( im->mll_free >= im->mll_max )
	      error(OVER_FLOWED,"parse_LINE 2");
	    mll = &im->mll[im->mll_free];
	    mll->mls_idx = im->mls_free;
	    mll->mpl_head = multi2list_new_head(im->mpl);
	    if( (len = parse_POS(im, p + sum, mll->mpl_head)) > 0 ){
	      sum += len;
	      while( (len = parse_POS(im, p + sum, mll->mpl_head)) > 0 ){
		sum += len;
	      }
	      if( (len = scan_RET(im->block, p + sum)) > 0 ){
		mll->cnt = CONT;
		im->mls_free += 1;
		im->mll_free += 1;
		return sum + len;
	      }
	    }
	  }
	}
      }
    }
  return 0;
}



int parse_BODY(InMorph *im, int p)
{
  int sum, len;

  if( (sum = parse_LINE(im, p)) > 0 ){
    while( (len = parse_LINE(im, p + sum)) > 0 ){
      sum += len;
    }
    if( (len = parse_LINENIL(im, p + sum)) > 0 ){
      return sum + len;
    }
  }
  return 0;
}



#define parse_SENTENCE(_im,_p) parse_HEADER((_im),(_p))

int parse_HEADER(InMorph *im, int p)
{
  int sum, len;

  if( (sum = scan_NOTRETS(im->block, p)) > 0 )
    if( (len = scan_RET(im->block, p + sum)) > 0 ){
      return sum + len;
    }
  return 0;
}



int parse_MORPH(InMorph *im, int p)
{
  int sum, len;

  if( (sum = parse_HEADER(im, p)) > 0 )
    if( (len = parse_SENTENCE(im, p + sum)) > 0 ){
      sum += len;
      if( (len = parse_BODY(im, p + sum)) > 0 ){
	return sum + len;
      }
    }
  return 0;
}
