#include <stdlib.h>
#include "error.h"
#include "map_linker.h"
#include "stype.h"
#include "atype.h"
#include "trie.h"
#include "keyword.h"

#define MAX_SYMBOL_BUFFER	10000	// ѿ̾10ʸ * 1000ѿ
#define MAX_SYMBOLS		1000	// ѿ 1000 

typedef struct {
  DefType type;
  int index;	// STYPEξ硢PreTermATYPEξ硢Network γƥǥå
} CndSym;

typedef struct {
  CndSym *cs;
  int max;
  int free;
} CndSymTable;

// ------------------------------------------------------------------

void cnd_symbol_table_clear(CndSymTable *cst)
{
  cst->free = 0;
}


CndSymTable *cnd_symbol_table_open(int size)
{
  CndSymTable *cst;

  if( (cst = (CndSymTable *)malloc(sizeof(CndSymTable))) == NULL )
    error(MALLOC_FAILED,"cnd_symbol_table_open 1");
  if( (cst->cs = (CndSym *)malloc(sizeof(CndSym) * size)) == NULL )
    error(MALLOC_FAILED,"cnd_symbol_table_open 2");
  cst->max = size;
  cnd_symbol_table_clear(cst);
  return cst;
}


void cnd_symbol_table_close(CndSymTable *cst)
{
  free(cst->cs);
  free(cst);
}


int cnd_symbol_table_set(CndSymTable *cst, DefType vt, int idx)
{
  if( cst->free >= cst->max )
    error(OVER_FLOWED,"cnd_symbol_table_set 1");
  cst->cs[cst->free].type = vt;
  cst->cs[cst->free].index = idx;
  return cst->free++;
}




// SޥåפϿƤФ򡢥ơ֥˥ȥꤹ
// ФʸϡSޥåפ StringBank Ȥ
void cnd_symbol_table_entry_stype(CndSymTable *cst, TrieC94 *ent, SType *st)
{
  int pt_size, i;
  char *w;

  pt_size = st->ptb->free;
  for( i = 0; i < pt_size; i += 1 ){
    w = string_bank_ref(st->sb, st->ptb->preterm[i].name);
    if( trie_c94_lookup(ent, w) < 0 ){
      trie_c94_entry(ent,w,cnd_symbol_table_set(cst, STYPE, i));
    } else {
      fprintf(stderr,"in STYPE: ");
      error(DUPLICATING_DECLARATION,w);
    }
  }
}

// AޥåפϿƤФ򡢥ơ֥˥ȥꤹ
// ФʸϡAޥåפ StringBank Ȥ
void cnd_symbol_table_entry_atype(CndSymTable *cst, TrieC94 *ent, AType *at, int start)
{
  int nt_size, i;
  char *w;

  nt_size = at->nb->free;
  for( i = start; i < nt_size; i += 1 ){
    w = string_bank_ref(at->sb, at->nb->network[i].name);
    if( trie_c94_lookup(ent, w) < 0 ){
      trie_c94_entry(ent,w,cnd_symbol_table_set(cst, ATYPE, i));
    } else {
      fprintf(stderr,"in ATYPE: ");
      error(DUPLICATING_DECLARATION,w);
    }
  }
}


// AξˤơVARFSYΥ٥CndSymTableΥǥå˽񤭴
void cnd_symbol_table_replace(CndSymTable *cst, TrieC94 *ent, AType *at)
{
  int i, idx;
  char *w;

  for( i = 0; i < at->cb->free; i += 1 ){
    switch( at->cb->cnd[i].ctype ){
    case VAR:
      w = string_bank_ref(at->sb, at->cb->cnd[i].str);
      idx = trie_c94_lookup(ent, w);
      if( idx < 0 )
	error(UNDEFINED_SYMBOL,w);
      switch( cst->cs[idx].type ){
      case ATYPE:
	at->cb->cnd[i].ctype = VARA;
	at->cb->cnd[i].link = cst->cs[idx].index;
	break;
      case STYPE:
	at->cb->cnd[i].ctype = VARS;
	at->cb->cnd[i].link = cst->cs[idx].index;
	break;
      }
      break;
    case FSY:
      w = string_bank_ref(at->sb, at->cb->cnd[i].str);
      idx = trie_c94_lookup(ent, w);
      if( idx < 0 )
	error(UNDEFINED_SYMBOL,w);
      switch( cst->cs[idx].type ){
      case ATYPE:
	error(UNKNOWN,"FSY  AϥݡȤƤޤ");
	break;
      case STYPE:
	at->cb->cnd[i].link = cst->cs[idx].index;
	break;
      }
      break;
    case RSN:
      w = string_bank_ref(at->sb, at->cb->cnd[i].str);
      idx = trie_c94_lookup(ent, w);
      if( idx < 0 )
	error(UNDEFINED_SYMBOL,w);
      switch( cst->cs[idx].type ){
      case ATYPE:
	at->cb->cnd[i].link = cst->cs[idx].index;
	break;
      case STYPE:
	error(UNKNOWN,"RSN  SϥݡȤƤޤ");
	break;
      }
      break;
    default:
      break;
    }
  }
}



// ɽ
void cnd_symbol_table_dump(CndSymTable *cst, FILE *fp)
{
  int i;

  for( i = 0; i < cst->free; i += 1 ){
    switch( cst->cs[i].type ){
    case STYPE:
      fprintf(fp,"%d:[S, %d]\n", i, cst->cs[i].index);
      break;
    case ATYPE:
      fprintf(fp,"%d:[A, %d]\n", i, cst->cs[i].index);
      break;
    }
  }
}


// ==================================================================
// SAѥΥޥåץե󥯤
// AѥΥޥåץեθˤäĤ

void map_linker(SType *st, AType *at, AType *pt)
{
  CndSymTable *cst;
  TrieC94 *ent;
  int offst_nb;

  ent = trie_c94_open(MAX_SYMBOL_BUFFER, ' ');
  cst = cnd_symbol_table_open(MAX_SYMBOLS);
  offst_nb = atype_cat_atype(pt, at); // pt θ at 򤯤äĤ

  cnd_symbol_table_entry_stype(cst, ent, st); // SθФϿ
  cnd_symbol_table_entry_atype(cst, ent, pt, offst_nb); // AθФϿ
  cnd_symbol_table_replace(cst, ent, pt); // VARFSYΥ񤭴
}
