/*
 * multilist.c
 * 2005.7/2 by tokuhisa
 *
 * ꥹΰ礭ȤꡢͭƻȤ
 * free 1ġhead ʣΩ
 * head ȯԡꥢʣ
 *       push/pop, enqueue/dequeue, car/cdr
 */

#include <stdio.h>
#include <stdlib.h>
#include "multilist.h"
#include "error.h"



void multilist_clear(MultiList *ml)
{
  int i;

  ml->free = 1;
  ml->node[0].next = 0;
  ml->node[0].val = 0;
  for( i = 1; i < ml->max; i += 1 )
    ml->node[i].next = i + 1;
  ml->node[i-1].next = 0;

#ifdef _TEST_FREE_AREA_
  ml->entnum = 0;
#endif

  //  printf("clear multilist free = %d\n", ml->free);
}



MultiList *multilist_open(int size)
{
  MultiList *ml;

  if( (ml = (MultiList *)malloc(sizeof(MultiList))) == NULL )
    error(MALLOC_FAILED,"multilist_open 1");
  if( (ml->node = (MLNode *)malloc(sizeof(MLNode) * size)) == NULL )
    error(MALLOC_FAILED,"multilist_open 2");
  ml->max = size;
  multilist_clear(ml);

  //  printf("open multilist free = %d\n", ml->free);
  return ml;
}



void multilist_close(MultiList *ml)
{
  free(ml->node);
  free(ml);
}


void multilist_dump(MultiList *ml, FILE *fp)
{
  MLNode *n;
  int i;

  for( i = 0; i < ml->max ; i += 1 ){
    n = &ml->node[i];
    fprintf(fp, "%3d:[%3d|%3d]",i,n->val, n->next);
    if( i == ml->free )
      fprintf(fp, " <-- free\n");
    else
      fprintf(fp, "\n");
  }
}



int multilist_new_head(MultiList *ml)
{
  int ans;

  if( ml->free == 0 )
    error(OVER_FLOWED,"multilist_new_head 1");

  ans = ml->free;
  ml->free = ml->node[ans].next;
  ml->node[ans].next = 0;
  ml->node[ans].val = ans;	// node[head].val  tail

#ifdef _TEST_FREE_AREA_
  ml->entnum += 1;
#endif

  //  printf("new head multilist free = %d\n", ml->free);
  return ans;
}


void multilist_clear_head(MultiList *ml, int head)
{
  int tail;

  tail = head;
  while( ml->node[tail].next != 0 ){
    tail = ml->node[tail].next;
#ifdef _TEST_FREE_AREA_
    ml->entnum -= 1;
#endif
  }
  ml->node[tail].next = ml->free;
  ml->free = ml->node[head].next;
  ml->node[head].next = 0;

  //  printf("clear head multilist free = %d\n", ml->free);
}




void multilist_close_head(MultiList *ml, int head)
{
  int tail;

  tail = head;
  while( ml->node[tail].next != 0 ){
    tail = ml->node[tail].next;
#ifdef _TEST_FREE_AREA_
    ml->entnum -= 1;
#endif

  }
  ml->node[tail].next = ml->free;
  ml->free = head;

#ifdef _TEST_FREE_AREA_
  ml->entnum -= 1;
#endif

  //  printf("close head multilist free = %d\n", ml->free);
}



void multilist_dump_head(MultiList *ml, int head, FILE *fp)
{
  int p;

  p = ml->node[head].next;
  while( p != 0 ){
    printf("[%d]",ml->node[p].val);
    p = ml->node[p].next;
  }
  printf("\n");
}



void multilist_enqueue(MultiList *ml, int head, int val)
{
  int t;

  if( ml->free == 0 )
    error(OVER_FLOWED,"multilist_enqueue");

  t = ml->free;
  ml->free = ml->node[t].next;

  ml->node[t].val = val;
  ml->node[t].next = 0;
  ml->node[ml->node[head].val].next = t;   // ml->node[head].val = tail
  ml->node[head].val = t;

#ifdef _TEST_FREE_AREA_
  ml->entnum += 1;
#endif

  //  printf("enqueue multilist free = %d\n", ml->free);
}



int multilist_dequeue(MultiList *ml, int head)
{
  int t;

  t = ml->node[head].next;
  ml->node[head].next = ml->node[t].next;
  if( t == ml->node[head].val )
    ml->node[head].val = head;
  ml->node[t].next = ml->free;
  ml->free = t;

#ifdef _TEST_FREE_AREA_
  ml->entnum -= 1;
#endif
  //  printf("dequeue multilist free = %d\n", ml->free);

  return ml->node[t].val;
}



void multilist_push(MultiList *ml, int head, int val)
{
  int t;

  if( ml->free == 0 )
    error(OVER_FLOWED,"multilist_push");

  t = ml->free;
  ml->free = ml->node[t].next;

  ml->node[t].val = val;
  ml->node[t].next = ml->node[head].next;
  ml->node[head].next = t;  

#ifdef _TEST_FREE_AREA_
  ml->entnum += 1;
#endif

  //  printf("push multilist free = %d\n", ml->free);
}


int multilist_pop(MultiList *ml, int head)
{
  int t;

  t = ml->node[head].next;
  ml->node[head].next = ml->node[t].next;
  ml->node[t].next = ml->free;
  ml->free = t;
  //  printf("pop multilist free = %d\n", ml->free);

#ifdef _TEST_FREE_AREA_
  ml->entnum -= 1;
#endif

  return ml->node[t].val;
}



void multilist_dup(MultiList *ml, int d, int s)
{
  int t;

  s = ml->node[s].next;
  while( s != 0 ){
    t = ml->free;			// ΰ
    if( t == 0 )
      error(OVER_FLOWED,"multilist_dup");
    ml->free = ml->node[t].next;
    ml->node[t].val = ml->node[s].val;	// 񤭹
    ml->node[t].next = 0;
    ml->node[d].next = t;		// ³
    d = t;				// 
    s = ml->node[s].next;
#ifdef _TEST_FREE_AREA_
    ml->entnum += 1;
#endif
  }

  //  printf("dup multilist free = %d\n", ml->free);
}


// val ξŤͽ񤭡
void multilist_insert_ow(MultiList *ml, int head, int val)
{
  int p,t,v;

  for( p = head; ml->node[p].next != 0; p = ml->node[p].next ){
    v = ml->node[ml->node[p].next].val;
    if( val < v )
      break;
    else if( val == v )
      return;
  }
  if( ml->free == 0 )
    error(OVER_FLOWED,"multilist_insert_ow");
  t = ml->free;
  ml->free = ml->node[t].next;
  ml->node[t].next = ml->node[p].next;
  ml->node[t].val = val;
  ml->node[p].next = t;
#ifdef _TEST_FREE_AREA_
  ml->entnum += 1;
#endif

}



// ==================================================================
// ==================================================================
// ==================================================================

void multi2list_clear(Multi2List *ml)
{
  int i;

  ml->free = 1;
  ml->node[0].next = 0;
  ml->node[0].val1 = 0;
  ml->node[0].val2 = 0;
  for( i = 1; i < ml->max; i += 1 )
    ml->node[i].next = i + 1;
  ml->node[i-1].next = 0;

#ifdef _TEST_FREE_AREA_
  ml->entnum = 0;
#endif

}



Multi2List *multi2list_open(int size)
{
  Multi2List *ml;

  if( (ml = (Multi2List *)malloc(sizeof(Multi2List))) == NULL )
    error(MALLOC_FAILED,"multi2list_open 1");
  if( (ml->node = (ML2Node *)malloc(sizeof(ML2Node) * size)) == NULL )
    error(MALLOC_FAILED,"multi2list_open 2");
  ml->max = size;
  multi2list_clear(ml);
  return ml;
}



void multi2list_close(Multi2List *ml)
{
  free(ml->node);
  free(ml);
}



int multi2list_new_head(Multi2List *ml)
{
  int ans;

  if( ml->free == 0 )
    error(OVER_FLOWED,"multi2list_new_head 1");

  ans = ml->free;
  ml->free = ml->node[ans].next;
  ml->node[ans].next = 0;
  ml->node[ans].val1 = ans;	// node[head].val1  tail

#ifdef _TEST_FREE_AREA_
  ml->entnum += 1;
#endif

  return ans;
}



void multi2list_close_head(Multi2List *ml, int head)
{
  int tail;

  tail = head;
  while( ml->node[tail].next != 0 ){
    tail = ml->node[tail].next;
#ifdef _TEST_FREE_AREA_
    ml->entnum -= 1;
#endif
  }
  ml->node[tail].next = ml->free;
  ml->free = head;

#ifdef _TEST_FREE_AREA_
  ml->entnum -= 1;
#endif

}



void multi2list_enqueue(Multi2List *ml, int head, int val1, int val2)
{
  int t;

  if( ml->free == 0 )
    error(OVER_FLOWED,"multi2list_enqueue");

  t = ml->free;
  ml->free = ml->node[t].next;

  ml->node[t].val1 = val1;
  ml->node[t].val2 = val2;
  ml->node[t].next = 0;
  ml->node[ml->node[head].val1].next = t;   // ml->node[head].val1 = tail
  ml->node[head].val1 = t;

#ifdef _TEST_FREE_AREA_
  ml->entnum += 1;
#endif
}



ML2Node *multi2list_dequeue(Multi2List *ml, int head)
{
  int t;

  t = ml->node[head].next;
  if( t == 0 )
    error(UNDER_FLOWED,"multi2lit_dequeue");

  ml->node[head].next = ml->node[t].next;
  if( t == ml->node[head].val1 )
    ml->node[head].val1 = head;
  ml->node[t].next = ml->free;
  ml->free = t;

#ifdef _TEST_FREE_AREA_
  ml->entnum -= 1;
#endif

  return &ml->node[t];
}



void multi2list_push(Multi2List *ml, int head, int val1, int val2)
{
  int t;

  if( ml->free == 0 )
    error(OVER_FLOWED,"multi2list_push");

  t = ml->free;
  ml->free = ml->node[t].next;

  ml->node[t].val1 = val1;
  ml->node[t].val2 = val2;
  ml->node[t].next = ml->node[head].next;
  ml->node[head].next = t;  
#ifdef _TEST_FREE_AREA_
  ml->entnum += 1;
#endif

}


ML2Node *multi2list_pop(Multi2List *ml, int head)
{
  int t;

  t = ml->node[head].next;
  if( t == 0 )
    error(UNDER_FLOWED,"multi2list_pop");
  ml->node[head].next = ml->node[t].next;
  ml->node[t].next = ml->free;
  ml->free = t;

#ifdef _TEST_FREE_AREA_
  ml->entnum -= 1;
#endif

  return &ml->node[t];
}



void multi2list_dup(Multi2List *ml, int d, int s)
{
  int t;

  s = ml->node[s].next;
  while( s != 0 ){
    t = ml->free;			// ΰ
    if( t == 0 )
      error(OVER_FLOWED,"multi2list_dup");
    ml->free = ml->node[t].next;
    ml->node[t].val1 = ml->node[s].val1;	// 񤭹
    ml->node[t].val2 = ml->node[s].val2;	// 񤭹
    ml->node[t].next = 0;
    ml->node[d].next = t;		// ³
    d = t;				// 
    s = ml->node[s].next;
#ifdef _TEST_FREE_AREA_
    ml->entnum += 1;
#endif
  }
}


ML2Node *multi2list_search_by_val1(Multi2List *ml, int head, int key)
{
  int p;

  p = ml->node[head].next;
  while( p != 0 ){
    if( ml->node[p].val1 == key )
      return &ml->node[p];
    p = ml->node[p].next;
  }
  return NULL;
}
