#include <stdlib.h>
#include "st.h"
#include "error.h"
#include "string_bank.h"

StringBank *string_bank_new(int lsize, int wsize)
{
  StringBank *ans;

  ans = (StringBank *)malloc(sizeof(StringBank));
  if( ans == NULL )
    error(MALLOC_FAILED,"string_bank_new 1");

  ans -> all_letter = (char *)malloc(sizeof(char) * lsize);
  if( ans -> all_letter == NULL )
    error(MALLOC_FAILED,"string_bank_new 2");
  ans->max_letter = lsize;
  ans->free_letter = 0;

  ans->all_word = (int *)malloc(sizeof(int) * wsize);
  if( ans->all_word == NULL )
    error(MALLOC_FAILED,"string_bank_new 3");
  ans->max_word = wsize;
  ans->free_word = 0;

  ans -> word_entry = st_init_strtable_with_size(wsize);

  return ans;
}


int string_bank_add(StringBank *sb, char *word)
{
  int len;
  st_data_t tmp;

  if( st_lookup(sb->word_entry, (st_data_t)word, (st_data_t *)&tmp) )
    return (int)tmp;

  if( sb->max_word == sb->free_word )
    error(OVER_FLOWED,"string_bank_add 1");

  len = strlen(word);
  if( len + sb->free_letter + 1 >= sb->max_letter )
    error(OVER_FLOWED,"string_bank_add 2");

  strcpy(&(sb->all_letter[sb->free_letter]), word);

  sb->all_word[sb->free_word] = sb->free_letter;
  st_insert(sb->word_entry, (st_data_t)&sb->all_letter[sb->free_letter], (st_data_t)sb->free_word);

  sb->free_letter += len + 1;
  return sb->free_word++;
}



int string_bank_add_n(StringBank *sb, char *word, int len)
{
  st_data_t tmp;

  if( sb->max_word == sb->free_word )
    error(OVER_FLOWED,"string_bank_add_n 1");

  if( len + sb->free_letter + 1 >= sb->max_letter )
    error(OVER_FLOWED,"string_bank_add_n 2");

  strncpy(&(sb->all_letter[sb->free_letter]), word, len);
  sb->all_letter[sb->free_letter + len] = '\0';

  if( st_lookup(sb->word_entry, (st_data_t)&sb->all_letter[sb->free_letter], (st_data_t *)&tmp) )
    return (int)tmp;

  sb->all_word[sb->free_word] = sb->free_letter;
  st_insert(sb->word_entry, (st_data_t)&sb->all_letter[sb->free_letter], (st_data_t)sb->free_word);

  sb->free_letter += len + 1;
  return sb->free_word++;
}


void string_bank_dump(StringBank *sb, FILE *fp)
{
  int i;

  fprintf(fp,"StringBank\n");
  fprintf(fp,"letter:  max = %d   free = %d\n", sb->max_letter, sb->free_letter);
  fprintf(fp, "word:  max = %d   free = %d\n", sb->max_word, sb->free_word);

  for( i = 0; i < sb->free_letter; i += 1 )
    if( sb->all_letter[i] == '\0' )
      fprintf(fp,"\\0");
    else
      fprintf(fp,"%c",sb->all_letter[i]);
  fprintf(fp,"\n");
  
  for( i = 0; i < sb->free_word; i += 1 )
    fprintf(fp,"%d:[%s]\n", i, &(sb->all_letter[sb->all_word[i]]));
}


char *string_bank_ref(StringBank *sb, int i)
{
  return &(sb->all_letter[sb->all_word[i]]);
}


void string_bank_save(StringBank *sb, FILE *fp)
{
  // -------------------- إå --------------------
  fprintf(fp,"%d %d %d %d\n",sb->max_letter,sb->free_letter,sb->max_word,sb->free_word);
  // -------------------- ܥǥ --------------------
  fwrite(sb->all_letter, sizeof(char), sb->free_letter, fp);
  fwrite(sb->all_word, sizeof(int), sb->free_word, fp);
}



StringBank *string_bank_load(FILE *fp)
{
  StringBank *ans;
  int i;
  st_data_t tmp;

  ans = (StringBank *)malloc(sizeof(StringBank));
  if( ans == NULL )
    error(MALLOC_FAILED,"string_bank_load 1");

  // -------------------- إå --------------------
  fscanf(fp,"%d %d %d %d\n",&ans->max_letter,&ans->free_letter,&ans->max_word,&ans->free_word);

  ans->all_letter = (char *)malloc(sizeof(char)*ans->max_letter);
  if( ans->all_letter == NULL )
    error(MALLOC_FAILED,"string_bank_load 2");

  ans->all_word = (int *)malloc(sizeof(int)*ans->max_word);
  if( ans->all_word == NULL )
    error(MALLOC_FAILED,"string_bank_load 3");

  // -------------------- ܥǥ --------------------
  fread(ans->all_letter, sizeof(char), ans->free_letter, fp);
  fread(ans->all_word, sizeof(int), ans->free_word, fp);


  // ϥåơ֥
  ans -> word_entry = st_init_strtable_with_size(ans->max_word);
  for( i = 0 ; i < ans->free_word; i += 1)
    if( !st_lookup(ans->word_entry, (st_data_t)&(ans->all_letter[ans->all_word[i]]), (st_data_t *)&tmp) )
      st_insert(ans->word_entry, (st_data_t)&(ans->all_letter[ans->all_word[i]]), (st_data_t)i);
  
  return ans;
}
