/*
 * trie.c
 * 2005.6/22 by tokuhisa
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "error.h"
#include "trie.h"

// ==================================================================
// 26ʸΥȥ饤

TrieC26 *trie_c26_open(int size, char base)
{
  TrieC26 *ans;

  if( (ans = (TrieC26 *)malloc(sizeof(TrieC26))) == NULL )
    error(MALLOC_FAILED,"trie_c26_open 1");
  if( (ans->node = (TrieC26Node *)malloc(sizeof(TrieC26Node) * size)) == NULL )
    error(MALLOC_FAILED,"trie_c26_open 2");
  ans->max = size;
  ans->base = base;
  trie_c26_clear(ans);
  return ans;
}



void trie_c26_clear(TrieC26 *tc)
{
  int i;
  memset(tc->node, 0, sizeof(TrieC26Node) * tc->max);
  for( i = 0; i < tc->max; i += 1 )
    tc->node[i].num = -1;
  tc->free = 1;
}



void trie_c26_close(TrieC26 *tc)
{
  free(tc->node);
  free(tc);
}



void trie_c26_entry(TrieC26 *tc, char *word, int data)
{
  int n;
  char *p;

  n = 0;
  for( p = word; *p != '\0'; p += 1 ){
    if( tc->node[n].next[*p - tc->base] == 0 ){
      if( tc->free >= tc->max )
	error(OVER_FLOWED, "trie_c26_entry 1");
      tc->node[n].next[*p - tc->base] = tc->free;
      n = tc->free;
      tc->free += 1;
    } else {
      n = tc->node[n].next[*p - tc->base];
    }
  }
  tc->node[n].num = data;
}



void trie_c26_entry_n(TrieC26 *tc, char *word, int len, int data)
{
  int n, l;
  char *p;

  n = 0;
  for( p = word, l = 0; l < len; p += 1, l += 1 ){
    if( tc->node[n].next[*p - tc->base] == 0 ){
      if( tc->free >= tc->max )
	error(OVER_FLOWED, "trie_c26_entry 1");
      tc->node[n].next[*p - tc->base] = tc->free;
      n = tc->free;
      tc->free += 1;
    } else {
      n = tc->node[n].next[*p - tc->base];
    }
  }
  tc->node[n].num = data;
}



int trie_c26_lookup(TrieC26 *tc, char *word)
{
  int n;
  char *p;

  n = 0;
  for( p = word; *p != '\0'; p += 1 ){
    if( (n = tc->node[n].next[*p - tc->base]) == 0 )
      return -1;
  }
  return tc->node[n].num;
}



int trie_c26_lookup_n(TrieC26 *tc, char *word, int len)
{
  int n, l;
  char *p;

  n = 0;
  for( p = word, l = 0; l < len; p += 1, l += 1 ){
    if( (n = tc->node[n].next[*p - tc->base]) == 0 )
      return -1;
  }
  return tc->node[n].num;
}



void trie_c26_dump(TrieC26 *tc, FILE *fp)
{
  int i,j;

  fprintf(fp,"TrieC26 (max = %d, base = '%c')\n", tc->max, tc->base);
  fprintf(fp, "    ");
  for( j = 0; j < 26 ; j += 1 )
    fprintf(fp," %c ", tc->base + j );
  fprintf(fp,"\n");
  for( i = 0; i < tc->free; i += 1 ){
    fprintf(fp, "%2d:[", i );
    for( j = 0; j < 26; j += 1 )
      fprintf(fp, "%2d ", tc->node[i].next[j]);
    fprintf(fp,"%2d]\n",tc->node[i].num);
  }
}

// ==================================================================
// 94ʸΥȥ饤

TrieC94 *trie_c94_open(int size, char base)
{
  TrieC94 *ans;

  if( (ans = (TrieC94 *)malloc(sizeof(TrieC94))) == NULL )
    error(MALLOC_FAILED,"trie_c94_open 1");
  if( (ans->node = (TrieC94Node *)malloc(sizeof(TrieC94Node) * size)) == NULL )
    error(MALLOC_FAILED,"trie_c94_open 2");
  ans->max = size;
  ans->base = base;
  trie_c94_clear(ans);
  return ans;
}



void trie_c94_clear(TrieC94 *tc)
{
  int i;
  memset(tc->node, 0, sizeof(TrieC94Node) * tc->max);
  for( i = 0; i < tc->max; i += 1 )
    tc->node[i].num = -1;
  tc->free = 1;
}



void trie_c94_close(TrieC94 *tc)
{
  free(tc->node);
  free(tc);
}



void trie_c94_entry(TrieC94 *tc, char *word, int data)
{
  int n;
  char *p;

  n = 0;
  for( p = word; *p != '\0'; p += 1 ){
    if( tc->node[n].next[*p - tc->base] == 0 ){
      if( tc->free >= tc->max )
	error(OVER_FLOWED, "trie_c94_entry 1");
      tc->node[n].next[*p - tc->base] = tc->free;
      n = tc->free;
      tc->free += 1;
    } else {
      n = tc->node[n].next[*p - tc->base];
    }
  }
  tc->node[n].num = data;
}



void trie_c94_entry_n(TrieC94 *tc, char *word, int len, int data)
{
  int n, l;
  char *p;

  n = 0;
  for( p = word, l = 0; l < len; p += 1, l += 1 ){
    if( tc->node[n].next[*p - tc->base] == 0 ){
      if( tc->free >= tc->max )
	error(OVER_FLOWED, "trie_c94_entry 1");
      tc->node[n].next[*p - tc->base] = tc->free;
      n = tc->free;
      tc->free += 1;
    } else {
      n = tc->node[n].next[*p - tc->base];
    }
  }
  tc->node[n].num = data;
}



int trie_c94_lookup(TrieC94 *tc, char *word)
{
  int n;
  char *p;

  n = 0;
  for( p = word; *p != '\0'; p += 1 ){
    if( (n = tc->node[n].next[*p - tc->base]) == 0 )
      return -1;
  }
  return tc->node[n].num;
}



int trie_c94_lookup_n(TrieC94 *tc, char *word, int len)
{
  int n, l;
  char *p;

  n = 0;
  for( p = word, l = 0; l < len; p += 1, l += 1 ){
    if( (n = tc->node[n].next[*p - tc->base]) == 0 )
      return -1;
  }
  return tc->node[n].num;
}



void trie_c94_dump(TrieC94 *tc, FILE *fp)
{
  int i,j;

  fprintf(fp,"TrieC94 (max = %d, base = '%c')\n", tc->max, tc->base);
  fprintf(fp, "    ");
  for( j = 0; j < 94 ; j += 1 )
    fprintf(fp," %c ", tc->base + j );
  fprintf(fp,"\n");
  for( i = 0; i < tc->free; i += 1 ){
    fprintf(fp, "%2d:[", i );
    for( j = 0; j < 94; j += 1 )
      fprintf(fp, "%2d ", tc->node[i].next[j]);
    fprintf(fp,"%2d]\n",tc->node[i].num);
  }
}
